part of '../export.dart';

/// A Wave with a synchronously available and changeable value. Meant to reduce Signal boilerplate
///
/// - It doesn't implement Sink to suppress the close-sinks lint warning. (A signal does)
///
/// - The & and > operators should be used to dispose this Sea.
///
/// - The top-level method [bye] is meant to be used with &:
///
/// ```
/// class MyBloc extends HookBloc {
///   final Sea<int> age = Sea(18) & bye;
/// ```
///
/// - The [SeaHookBlocExtension.b] method is supposed to be used with the > operator
/// like this:
/// ```
/// class MyBloc extends HookBloc {
///   ...
///   MyBloc() {
///     ... = Sea(1) > b;
///   }
/// }
/// ```
///
///
///
/// [Sea] is also a [Wave] so that it's backwards compatible with consumers
/// expecting a [Wave].
///
/// Add is only meant to be called inside of HookBlocs. That's why it's set
/// as protected.
class Sea<V> extends Wave<V> with WaveType<V>, WaveOtherMixin<V>, WaveOperators<V> {
  final Future<void> Function() _close;
  final Wave<V> wave;

  /// .add is supposed to be called in HookBloc subclasses never outside.
  /// You're supposed to add this line to the top of your HookBloc
  // ignore_for_file: invalid_use_of_protected_member
  @protected
  final void Function(V) add;
  final V Function() __value;

  factory Sea(V s) {
    final signal = Signal(s);
    return Sea._create(
          () async => signal.close(),
      signal.wave,
      signal.add,
          () => signal.value,
    );
  }

  Sea._create(this._close, this.wave, this.add, this.__value) : super(wave);

  @override
  V get value => __value();

  Sea<V> operator &(Sea<V> Function(Sea<V> self, void Function() dispose) action) {
    return action(this, _close);
  }

  void change(V Function(V) action) => action(value);
}

/// extension operators for convenient disposal when used with HookBlocs.
T bye<T extends Sea<dynamic>>(T t, void Function() dispose) {
  return HookBloc.disposeEffect(t, (a) async => dispose());
}

/// Experimental High level HookBloc extensions for boilerplate free Blocs.
extension SeaHookBlocExtension on HookBloc {
  T b<T>(T t, void Function() dispose) {
    onDispose.add(() async => dispose());
    return t;
  }
}
